#!/usr/bin/perl -w
#
# Script zum ersetzen einer Zeichenfolge in Dateinamen
# und für sonstige Umbenennungen.
#
# Version 1.5
#
# Copyright (C) 2003-2009 Helmar Gerloni
#
# Author: Helmar Gerloni <helmar@gerloni.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# http://www.gnu.org/licenses/gpl.txt
#

# Module (siehe man 3pm Modulname)
use Getopt::Long;
use File::Basename;
use File::stat;
use Term::ANSIColor;

# Globale Variablen
#  Kommandoname
$cmd = basename($0);
#  Die Optionen werden von GetOptions aus @ARGV entfernt
leseoptionen();
#  Zu durchsuchende Verzeichnisse
@dir = @ARGV;
#  Trennzeichen
$trenn1 = "_-_";
$trenn2 = "__";
#  Sonstige Variablen
$dir = '';
$oldname = '';
$newname = '';

# Optionen einlesen und Warnungen abfangen
sub leseoptionen {
  $new = '';
  local $SIG{__WARN__} = sub { my $warnung; $warnung=$_[0]; chomp($warnung); AufrufFehler($warnung); };
  GetOptions ('help' => \$help, 'old=s' => \$old, 'iold=s' => \$iold, 'new=s' => \$new, 'test' => \$tst, 'uc1' => \$uc1, 'enum' => \$enum, 'ucall' => \$ucall, 'lcall' => \$lcall, 'oname=s' => \$oname, 'nname=s' => \$nname, 'swap' => \$swap);
}

# --help überschreibt alle anderen Optionen
if ($help) {
  Hilfe();
  exit;
}

# Aufruf untersuchen und entsprechende Funktionen aufrufen
if (@dir && $old && !$iold && !$uc1 && !$enum && !$ucall && !$lcall && !$oname  && !$nname && !$swap) {
  # Aufruf: --old [--new] dirs
  intro();
  if ($new) {
    print "- Ersetze in allen Dateinamen $old durch $new\n";
  } else {
    print "- Lösche aus allen Dateinamen $old\n";
  }
  linie();
  hauptfunktion();
} elsif (@dir && !$old && $iold && !$uc1 && !$enum && !$ucall && !$lcall && !$oname  && !$nname && !$swap) {
  # Aufruf: --iold [--new] dirs
  intro();
  if ($new) {
    print "- Ersetze in allen Dateinamen $iold durch $new (ignoriere Groß-/Kleinschreibung)\n";
  } else {
    print "- Lösche aus allen Dateinamen $iold (ignoriere Groß-/Kleinschreibung)\n";
  }
  linie();
  hauptfunktion();
} elsif (@dir && !$old && !$iold && !$new && $uc1 && !$enum && !$ucall && !$lcall && !$oname  && !$nname && !$swap) {
  # Aufruf: --uc1 dirs
  intro();
  print "- Mache den ersten Buchstaben jedes Dateinamens sowie nach $trenn1 und $trenn2 zum Großbuchstaben.\n";
  linie();
  hauptfunktion();
} elsif (@dir && !$old && !$iold && !$new && !$uc1 && !$enum && $ucall && !$lcall && !$oname  && !$nname && !$swap) {
  # Aufruf: --ucall dirs
  intro();
  print "- Mache alle Buchstaben jedes Dateinamens zu Großbuchstaben.\n";
  linie();
  hauptfunktion();
} elsif (@dir && !$old && !$iold && !$new && !$uc1 && !$enum && !$ucall && $lcall && !$oname  && !$nname && !$swap) {
  # Aufruf: --lcall dirs
  intro();
  print "- Mache alle Buchstaben jedes Dateinamens zu Kleinbuchstaben.\n";
  linie();
  hauptfunktion();
} elsif (@dir && !$old && !$iold && !$new && !$uc1 && !$enum && !$ucall && !$lcall && !$oname  && !$nname && $swap) {
  # Aufruf: --swap dirs
  intro();
  print "- Vertausche die Namensteile vor und nach dem ersten $trenn1 .\n";
  linie();
  hauptfunktion();
} elsif (@dir && !$old && !$iold && !$new && !$uc1 && $enum && !$ucall && !$lcall && !$oname  && !$nname && !$swap) {
  # Aufruf: --enum files
  intro();
  print "- Numeriere alle angegebenen Dateien zweistellig durch.\n";
  linie();
  enumer();
} elsif (!@dir && !$old && !$iold && !$new && !$uc1 && !$enum && !$ucall && !$lcall && $oname  && $nname && !$swap) {
  # Aufruf: --oname=list1 --nname=list2
  intro();
  print "- Benenne alle Dateien in $oname um in die Namen in $nname.\n";
  linie();
  onamenname();
} else {
  AufrufFehler("Ungültige Optionen-Kombination");
}

sub intro {
  linie();
  if ($tst) {
    print "- Testdurchlauf, es werden keine Veränderungen vorgenommen.\n";
    linie();
  }
}

sub mv_i {

  my $dir_oldname = '';
  my $dir_newname = '';

  if ($dir) {
    $dir_oldname = "$dir" . "/" . "$oldname";
    $dir_newname = "$dir" . "/" . "$newname";
    $dir_oldname =~ s/\/+/\//g;
    $dir_newname =~ s/\/+/\//g;
  } else {
    $dir_oldname = $oldname;
    $dir_newname = $newname;
  }

  # Abbruchbedingungen
  #  Keine Namensaenderung
  if ($newname eq $oldname) {
    return;
  }
  #  Alter Name leer
  if (!$oldname) {
    print color("red"), "  Alter Name ist leer.\n", color("reset");
    return;
  }
  #  Alter Name nicht vorhanden
  if (!-e $dir_oldname) {
    print color("red"), "  $oldname nicht vorhanden.\n", color("reset");
    return;
  }
  #  Neuer Name leer
  if (!$newname) {
    print color("red"), "  $oldname nicht umbenannt, neuer Name ist leer.\n", color("reset");
    return;
  }
  #  Neuer Name schon vorhanden
  if (-e $dir_newname) {
    print color("red"), "  $oldname nicht umbenannt, $newname in $dir schon vorhanden.\n", color("reset");
    return;
  }
  # Dateinamen, die mit - beginnen sind sehr schwer ansprechbar
    if (substr($newname, 0, 1) eq "-") {
    print color("red"), "  $oldname nicht umbenannt, $newname beginnt mit - .\n", color("reset");
    return;
  }
  # Wenn der neue Name / enthält, wird nur umbenannt, wenn das neue Verzeichnis auch existiert
  if ($newname =~ /\// && !-e dirname($dir_newname)) {
    print color("red"), "  $oldname nicht umbenannt, Verzeichnis ",dirname($dir_newname)," existiert nicht.\n", color("reset");
    return;
  }

  # Alles OK, aber nur Test
  if ($tst) {
    print "Test:  $oldname -> $newname\n";
    return;
  }

  # Alles OK, umbenennen
  print "  $oldname -> $newname\n";
  rename ("$dir_oldname", "$dir_newname") or die color("red"), "Unvorhergesehener Fehler beim Umbenennen von $oldname in $newname: $!\n", color("reset");

}

sub hauptfunktion {

  my @filesindir = ();
  my $file = '';
  my $endung = '';
  my $nummer = '';
  my @tmp = ();

  # Alle Verzeichnisse $dir in @dir
  VER: foreach $dir (@dir) {

    # Abbruchbedingungen
    #  dir nicht vorhanden
    if (!-e $dir) {
      print color("red"), "- Verzeichnis $dir nicht gefunden.\n", color("reset");
      next VER;
    }
    #  dir kein Verzeichnis
    if (!-d $dir) {
      # Meldung nur bei --test
      if ($tst) {
        print color("red"), "- $dir ist kein Verzeichnis.\n", color("reset");
      }
      next VER;
    }

    # Liste aller (alten) Dateinamen in $dir ausser . und .. vor dem Umbenennen in @filesindir schreiben
    @filesindir = ();
    opendir(DIR, $dir) or die color("red"), "Unvorhergesehener Fehler beim Öffnen von $dir: $!\n", color("reset");
    while ( defined ($file = readdir(DIR)) ) {
      if ($file ne "." && $file ne "..") {
        $#filesindir++;
        $filesindir[$#filesindir] = $file;
      }
    }
    closedir(DIR);

    # Alle Dateien @filesindir in $dir bearbeiten
    print color("green"), "- Verzeichnis $dir (", $#filesindir+1, " Dateien):\n", color("reset");
    foreach $oldname (@filesindir) {

      # newname ermitteln
      if ($uc1) {
        # Aufruf: --uc1 dirs
        #  Ersetzung nach trenn1
        @tmp = split(/$trenn1/,$oldname);
        foreach (@tmp) {
          $_ = ucfirst();
        }
        $newname = join($trenn1, @tmp);
        #  Ersetzung nach trenn2
        @tmp = split(/$trenn2/,$newname);
        foreach (@tmp) {
          $_ = ucfirst();
        }
        $newname = join($trenn2, @tmp);
      } elsif ($ucall) {
        # Aufruf: --ucall dirs
        $newname = uc($oldname);
      } elsif ($lcall) {
        # Aufruf: --lcall dirs
        $newname = lc($oldname);
      } elsif ($old) {
        # Aufruf: --old [--new] dirs
        ($newname = $oldname) =~ s/$old/$new/g;
      } elsif ($iold) {
        # Aufruf: --iold [--new] dirs
        ($newname = $oldname) =~ s/$iold/$new/ig;
      } elsif ($swap) {
        # Aufruf: --swap dirs
        if ($oldname =~ /$trenn1/) {
          @tmp = split(/$trenn1/,$oldname,2);
          # Vorderer Teil: Nummer(n) abschneiden, wenn vorhanden
          ($nummer, $tmp[0]) = $tmp[0] =~ /^([0-9]+[-_ ]?|)(.*)/;
          # Hinterer Teil: Endung abschneiden, wenn vorhanden
          if ($tmp[1] =~ /(.*)(\..+)/) {
            ($tmp[1], $endung) = $tmp[1] =~ /(.*)(\..+)/;
          } else {
            $endung = '';
          }
          $newname = $nummer.$tmp[1].$trenn1.$tmp[0].$endung;
        } else {
          $newname = $oldname;
        }
      } else {
        print color("red"), "- Interner Fehler!!!\n", color("reset");
        exit(1);
      }

      # umbenennen
      mv_i();
    }
  }
}

sub enumer {

  my $nr = "01";
  my ($tmp, $endung0) = $dir[0] =~ /^(.+)\.(.+)/;
  my $endungi = '';
  my $fileendok = -1;
  my $gleich = 0;

  # Test, ob alle angegebenen Dateien existieren, ob sie alle die gleiche
  # oder gar keine Endung (.xyz) haben und ob eine Datei doppelt angegeben wurde.
  if (!$endung0) {
    $endung0 = '';
  }
  foreach $oldname (@dir) {
    if (!-e $oldname) {
      print color("red"), "- Datei $oldname nicht gefunden.\n", color("reset");
      exit(1);
    }
    $gleich = 0;
    foreach $tmp (@dir) {
      if ($tmp eq $oldname) {
        $gleich++;
      }
      if ($gleich > 1) {
        print color("red"), "- Die Datei $oldname wurde doppelt angegeben.\n", color("reset");
        exit(1);
      }
    }
    ($tmp, $endungi) = $oldname =~ /^(.+)\.(.+)/;
    if (!$endungi) {
      $endungi = '';
    }
    if ($endungi ne $endung0) {
      print color("red"), "- Die angegebenen Dateien haben nicht alle dieselbe Endung.\n", color("reset");
      exit(1);
    }
  }

  # Test, ob die Anzahl der angegebenen Dateien.xyz (bzw. ohne Endung) gleich der Anzahl der
  # Dateien.xyz (bzw. ohne Endung) im aktuellen Verzeichnis ist (ob keine Datei vergessen wurde)
  opendir(DIR, ".") or die color("red"), "Unvorhergesehener Fehler beim Öffnen des aktuellen Verzeichnisses: $!\n", color("reset");
  if ($endung0) {
    while ( defined ($file = readdir(DIR)) ) {
      if ($file =~ /^(.+)\.$endung0$/) {
        $fileendok++;
      }
    }
  } else {
    while ( defined ($file = readdir(DIR)) ) {
      if ($file ne "." && $file ne ".." && !($file =~ /^(.+)\.(.+)/) ) {
        $fileendok++;
      }
    }
  }
  closedir(DIR);
  if ($fileendok ne $#dir) {
    print color("red"), "- Die Anzahl der angegebenen Dateien stimmt nicht mit der Anzahl der\n";
    if ($endung0) {
      print             "  .$endung0-Dateien im aktuellen Verzeichnis überein.\n", color("reset");
    } else {
      print             "  Dateien ohne Endung im aktuellen Verzeichnis überein.\n", color("reset");
    }
    exit(1);
  }

  # Alle Dateien $oldname in @dir umbenennen
  foreach $oldname (@dir) {
    $newname = "$nr" . "-" . "$oldname";

    # umbenennen
    mv_i();

    $nr++;
  }
}

sub onamenname {

  my @list1 = ();
  my @list2 = ();
  my $i = 0;

  # oname vorhanden?
  if (!-e $oname) {
    print color("red"), "- Datei $oname nicht gefunden.\n", color("reset");
    exit(1);
  }
  # nname vorhanden?
  if (!-e $nname) {
    print color("red"), "- Datei $nname nicht gefunden.\n", color("reset");
    exit(1);
  }

  # Dateien $oname und $nname einlesen und Zeilem zählen
  open(FILE, $oname) or die color("red"), "Unvorhergesehener Fehler beim Öffnen von $oname: $!\n", color("reset");
  while (<FILE>) {
    next if /^\s*(#|$)/;
    chomp;
    $#list1++;
    $list1[$#list1] = $_;
  }
  close FILE;
  open(FILE, $nname) or die color("red"), "Unvorhergesehener Fehler beim Öffnen von $nname: $!\n", color("reset");
  while (<FILE>) {
    next if /^\s*(#|$)/;
    chomp;
    $#list2++;
    $list2[$#list2] = $_;
  }
  close FILE;

  # Dateien müssen gleichviele Namen enthalten
  if ($#list1 ne $#list2) {
    print color("red"), "Dateien $oname (",$#list1+1," Zeilen) und $nname (",$#list2+1," Zeilen) sind nicht gleich lang.\n", color("reset");
    exit(1);
  }

  # Umbenennen
  for ($i = 0; $i <= $#list1; $i++) {
    $oldname = $list1[$i];
    $newname = $list2[$i];
    mv_i();
  }
}

sub linie {
  print "--------------------------------------------------------------------------------\n";
}

# Fehlerhafter Aufruf
sub AufrufFehler {
  linie();
  print "      Script zum Bearbeiten von Dateinamen.\n\n";
  print "      ",color("red"),"Fehlerhafter Aufruf:",color("reset"),"\n";
  print "        - $_[0].\n\n";
  print "      ",color("blue"),"$cmd --help",color("reset")," zeigt weitere Informationen an.\n";
  linie();
  exit(1);
}

# Kurzanleitung
sub Hilfe {
  linie();
  print "    Script zum Bearbeiten von Dateinamen.\n\n";
  print color("blue"),"  $cmd [--test] --[i]old=",color("green"),"old ",color("blue"),"[--new=",color("green"),"new",color("blue"),"] ",color("green"),"dir1 ",color("blue"),"[",color("green"),"dir2 ",color("blue"),"...]", color("reset"),"\n\n";
  print "    Ersetzt in den Verzeichnissen ",color("green"),"dir1 ",color("reset"),"[",color("green"),"dir2 ",color("reset"),"...] in allen Dateinamen\n";
  print "    ",color("green"),"old",color("reset")," durch ",color("green"),"new",color("reset"),".\n";
  print "    ",color("blue"),"--old",color("reset")," berücksichtigt Groß-/Kleinschreibung, ",color("blue"),"--iold=",color("reset")," nicht.\n";
  print "    Ohne Angabe von ",color("blue"),"--new=",color("green"),"new",color("reset")," wird ",color("green"),"old",color("reset")," aus allen Dateinamen gelöscht.\n";
  print "    ",color("green"),"old",color("reset")," kann reguläre Ausdrücke enthalten (",color("green"),"new",color("reset")," nicht).\n\n";
  print color("blue"), "  $cmd [--test] --uc1 ",color("green"),"dir1 ",color("blue"),"[",color("green"),"dir2",color("blue")," ...]\n\n", color("reset");
  print "    Macht den ersten Buchstaben jedes Dateinamens sowie nach $trenn1\n";
  print "    und $trenn2 in den Verzeichnissen ",color("green"),"dir1 ",color("reset"),"[",color("green"),"dir2",color("reset")," ...] zum Großbuchstaben.\n\n";
  print color("blue"), "  $cmd [--test] --ucall ",color("green"),"dir1 ",color("blue"),"[",color("green"),"dir2",color("blue")," ...]\n";
  print color("blue"), "  $cmd [--test] --lcall ",color("green"),"dir1 ",color("blue"),"[",color("green"),"dir2",color("blue")," ...]",color("reset"),"\n\n";
  print "    Macht alle Buchstaben jedes Dateinamens zum Groß- bzw. Kleinbuchstaben.\n\n";
  print color("blue"), "  $cmd [--test] --swap ",color("green"),"dir1 ",color("blue"),"[",color("green"),"dir2",color("blue")," ...]\n\n", color("reset");
  print "    Vertauscht die Namensteile vor und nach dem ersten $trenn1 aller Dateinamen\n";
  print "    in den Verzeichnissen ",color("green"),"dir1 ",color("reset"),"[",color("green"),"dir2",color("reset")," ...]. Führende Numerierung (z.B. 01-...)\n";
  print "    und Dateiendung werden beibehalten.\n\n";
  print color("blue"), "  $cmd [--test] --enum ",color("green"),"file1 ",color("blue"),"[",color("green"),"file2",color("blue")," ...]\n\n", color("reset");
  print "    Numeriert alle angegebenen Dateien/Verzeichnisse ",color("green"),"file1",color("reset")," [",color("green"),"file2",color("reset")," ...]\n";
  print "    nach dem Muster 01-",color("green"),"file1",color("reset"),", 02-",color("green"),"file2",color("reset")," ... durch.\n\n";
  print color("blue"), "  $cmd [--test] --oname=",color("green"),"list1",color("blue")," --nname=",color("green"),"list2",color("reset"),"\n\n";
  print "    ",color("green"),"list1",color("reset"),":       ",color("green"),"list2",color("reset"),":    Benennt alle Dateien aus der Liste in Datei ",color("green"),"list1",color("reset"),"\n";
  print "    ",color("red"),"oname1       nname1",color("reset"),"    um in die neuen Namen in Datei ",color("green"),"list2",color("reset"),".\n";
  print "    ",color("red"),"oname2       oname2\n";
  print "    ...          ...",color("reset"),"\n\n";
  print "  Die Option ",color("blue"),"--test",color("reset")," startet einen Testdurchlauf. Es wird angezeigt, was\n";
  print "  gemacht würde, es werden aber keine Änderungen vorgenommen. Außerdem werden\n";
  print "  etwas mehr Meldungen ausgegeben.\n";
  linie();
}

